@@ -42,12 +42,12 @@ using Gendarme.Framework.Rocks;
 namespace Gendarme.Rules.Naming {
 
 	/// <summary>
-	/// This rule ensure that identifier are correctly cased. In particular:
+	/// This rule ensures that identifiers are correctly cased. In particular:
 	/// <list>
-	/// <item><description>namespace names are pascal cased</description></item>
-	/// <item><description>type names are pascal cased</description></item>
-	/// <item><description>method names are pascal cased</description></item>
-	/// <item><description>parameter names are camel cased</description></item>
+	/// <item><description>namespace names are PascalCased</description></item>
+	/// <item><description>type names are PascalCased</description></item>
+	/// <item><description>method names are PascalCased</description></item>
+	/// <item><description>parameter names are camelCased</description></item>
 	/// </list>
 	/// </summary>
 	/// <example>
@@ -71,8 +71,8 @@ namespace Gendarme.Rules.Naming {
 	/// </code>
 	/// </example>
 
-	[Problem ("This identifier (namespace, type or method) violates the .NET naming conventions.")]
-	[Solution ("Change namespace, type and method names to be pascal-cased (like MyClass) and parameter names to be camel-cased (like myParameter).")]
+	[Problem ("This identifier (namespace, type, or method) violates the .NET naming conventions.")]
+	[Solution ("Change the namespace, type, or method name to be pascal-cased (like MyClass) and parameter names to be camel-cased (like myParameter).")]
 	[EngineDependency (typeof (NamespaceEngine))]
 	[FxCopCompatibility ("Microsoft.Naming", "CA1709:IdentifiersShouldBeCasedCorrectly")]
 	public class UseCorrectCasingRule : Rule, IAssemblyRule, ITypeRule, IMethodRule {
@@ -96,7 +96,7 @@ namespace Gendarme.Rules.Naming {
 				return name.ToUpperInvariant ();
 
 			int index = IndexOfFirstCorrectChar (name);
-			return Char.ToUpperInvariant (name [index]) + name.Substring (index + 1);
+			return Char.ToUpperInvariant (name [index]).ToString () + name.Substring (index + 1);
 		}
 
 		// check if name is camelCased
@@ -118,7 +118,7 @@ namespace Gendarme.Rules.Naming {
 				return name.ToLowerInvariant ();
 
 			int index = IndexOfFirstCorrectChar (name);
-			return Char.ToLowerInvariant (name [index]) + name.Substring (index + 1);
+			return Char.ToLowerInvariant (name [index]).ToString () + name.Substring (index + 1);
 		}
 
 		private static int IndexOfFirstCorrectChar (string s)
@@ -140,10 +140,11 @@ namespace Gendarme.Rules.Naming {
 			if (String.IsNullOrEmpty (nspace))
 				return;
 
+			NamespaceDefinition nd = NamespaceDefinition.GetDefinition (nspace);
 			foreach (string ns in nspace.Split ('.')) {
 				switch (ns.Length) {
 				case 1:
-					ReportCasingError (new NamespaceDefinition (nspace), string.Format (
+					ReportCasingError (nd, string.Format (
 						"Use of single character namespace is discouraged. Rename namespace {0}", ns));
 
 					break;
@@ -152,19 +153,19 @@ namespace Gendarme.Rules.Naming {
 					if (ns.Any (c => Char.IsLetter (c) && Char.IsLower (c))) {
 						string msg = String.Format ("In namespaces made of two characters, both characters should uppercase. Rename namespace '{0}' to '{1}'",
 							ns, ns.ToUpperInvariant ());
-						ReportCasingError (new NamespaceDefinition (nspace), msg);
+						ReportCasingError (nd, msg);
 					}
 					break;
 				default:
 					// if the sub namespace is made of 3 or more chars, make sure they're not all uppercase
 					if (ns.All (c => Char.IsLetter (c) && Char.IsUpper (c))) {
-						string msg = String.Format ("Namespaces longer than two characters should not be all uppercase. Rename namespace '{0}' to '{1}'",
-							ns, ns [0] + ns.Substring (1).ToLowerInvariant ());
-						ReportCasingError (new NamespaceDefinition (nspace), msg);
+						string msg = String.Format ("Namespaces longer than two characters should not be all uppercase. Rename namespace '{0}' to '{1}{2}'",
+							ns, ns [0].ToString (), ns.Substring (1).ToLowerInvariant ());
+						ReportCasingError (nd, msg);
 					} else if (!IsPascalCase (ns)) {
 						string msg = String.Format ("Namespaces longer than two characters should be pascal cased. Rename namespace '{0}' to '{1}'",
 							ns, PascalCase (ns));
-						ReportCasingError (new NamespaceDefinition (nspace), msg);
+						ReportCasingError (nd, msg);
 					}
 					break;
 				}
@@ -186,10 +187,12 @@ namespace Gendarme.Rules.Naming {
 				return RuleResult.DoesNotApply;
 
 			// types should all be PascalCased
-			if (!IsPascalCase (type.Name))
+			string name = type.Name;
+			if (!IsPascalCase (name)) {
 				ReportCasingError (type, string.Format (
-					"Type names should all be pascal-cased. Rename '{0}' type to '{1}'.", type.Name, PascalCase (type.Name)));
-
+					"Type names should all be pascal-cased. Rename '{0}' type to '{1}'.", 
+					name, PascalCase (name)));
+			}
 			return Runner.CurrentRuleResult;
 		}
 
@@ -202,15 +205,15 @@ namespace Gendarme.Rules.Naming {
 			if (method.IsConstructor || method.IsGeneratedCode ())
 				return RuleResult.DoesNotApply;
 
-			// don't consider private compiler generated add / remove on events
-			if ((method.IsAddOn || method.IsRemoveOn) && method.IsSynchronized && method.IsPrivate)
+			// don't consider private add / remove on events
+			if ((method.IsAddOn || method.IsRemoveOn) && method.IsPrivate)
 				return RuleResult.DoesNotApply;
 
 			string name = method.Name;
 			MethodSemanticsAttributes attrs = method.SemanticsAttributes;
 			if ((attrs & mask) != 0) {
 				// it's something special
-				int underscore = method.Name.IndexOf ('_');
+				int underscore = name.IndexOf ('_');
 				if (underscore != -1)
 					name = name.Substring (underscore + 1);
 			} else if (method.IsSpecialName) {
