@@ -27,6 +27,7 @@
 //
 
 using System;
+using System.Collections.Generic;
 
 using Mono.Cecil;
 using Mono.Cecil.Cil;
@@ -40,17 +41,25 @@ namespace Gendarme.Rules.Performance {
 
 	// rule request from
 	// https://bugzilla.novell.com/show_bug.cgi?id=406889
-
+	
 	/// <summary>
-	/// This rule looks for calls into <c>String</c> methods that use <b>String</b>
+	/// This rule looks for calls to <c>String</c> methods that use <b>String</b>
 	/// parameters when a <c>Char</c> parameter could have been used. Using the
-	/// <c>Char</c>-based method overload (ordinal comparison) is faster than the 
-	/// <c>String</c>-based one (culture-sensitive comparison). 
-	/// Note that .NET 4 will do a breaking change and will use ordinal comparison 
-	/// by default on <c>String</c>. If you need a culture-sensitive comparison 
-	/// use an overload that allow you to specify a <c>StringComparison</c>.
-	/// This rule checks specifically for <c>IndexOf</c>, <c>LastIndexOf</c> and
-	/// <c>Replace</c> to ensure there is a performance gain (other methods could vary).
+	/// <c>Char</c> overload is preferred because it will be faster. 
+	///
+	/// Note, however, that this may result in subtly different behavior on versions of
+	/// .NET before 4.0: the string overloads do a culture based comparison using
+	/// <c>CultureInfo.CurrentCulture</c> and the char methods do an ordinal
+	/// comparison (a simple compare of the character values). This can result in
+	/// a change of behavior (for example the two can produce different results when
+	/// precomposed characters are used). If this is important it is best to use an
+	/// overload that allows StringComparison or CultureInfo to be explicitly specified
+	/// see [http://msdn.microsoft.com/en-us/library/ms973919.aspx#stringsinnet20_topic4] 
+	/// for more details.
+	///
+	/// With .NET 4.0 <c>String</c>'s behavior will change and the various methods
+	/// will be made more consistent. In particular the comparison methods will be changed
+	/// so that they all default to doing an ordinal comparison.
 	/// </summary>
 	/// <example>
 	/// Bad example:
@@ -70,8 +79,8 @@ namespace Gendarme.Rules.Performance {
 	/// </example>
 	/// <remarks>This rule is available since Gendarme 2.4</remarks>
 
-	[Problem ("This code is calling string-based overload when char-based overload could be used.")]
-	[Solution ("Replace the single-sized string parameters with chararacter parameters.")]
+	[Problem ("This code is calling a string-based overload when a char-based overload could be used.")]
+	[Solution ("Replace the string parameters with chararacter parameters.")]
 	[EngineDependency (typeof (OpCodeEngine))]
 	public class PreferCharOverloadRule : Rule, IMethodRule {
 
@@ -82,21 +91,25 @@ namespace Gendarme.Rules.Performance {
 			return (parameter.Operand as string);
 		}
 
+		void Report (MethodDefinition method, Instruction ins, Confidence confidence, MemberReference call, string parameter)
+		{
+			string msg = String.Format ("Prefer the use of: {0}('{1}'...);", call.Name, parameter);
+			Runner.Report (method, ins, Severity.Medium, confidence, msg);
+		}
+
 		void CheckIndexOf (MethodDefinition method, MethodReference call, Instruction ins)
 		{
-			string msg;
-
 			// check that first parameter is a string of length equal to one
 			string p1 = GetString (ins.TraceBack (method, -1));
 			if (p1.Length != 1)
 				return;
 
-			int last = call.Parameters.Count;
-			if (call.Parameters [last - 1].ParameterType.FullName != "System.StringComparison") {
+			IList<ParameterDefinition> pdc = call.Parameters;
+			int last = pdc.Count;
+			if (pdc [last - 1].ParameterType.FullName != "System.StringComparison") {
 				// confidence is normal because it's possible that the code expects a
 				// culture sensitive comparison (but that will break in .NET 4).
-				msg = String.Format ("Prefer the use of: {0}('{1}'...);", call.Name, p1);
-				Runner.Report (method, ins, Severity.Medium, Confidence.Normal, msg);
+				Report (method, ins, Confidence.Normal, call, p1);
 				return;
 			}
 
@@ -105,14 +118,12 @@ namespace Gendarme.Rules.Performance {
 			switch (sc.OpCode.Code) {
 			case Code.Ldc_I4_4:
 				// if it's StringComparison.Ordinal (4) then it's identical to what a Char would do
-				msg = String.Format ("Prefer the use of: {0}('{1}'...);", call.Name, p1);
-				Runner.Report (method, ins, Severity.Medium, Confidence.High, msg);
+				Report (method, ins, Confidence.High, call, p1);
 				break;
 			case Code.Ldc_I4_5:
 				// if it's StringComparison.OrdinalIgnoreCase (5) then it's identical as long as the Char is not case sensitive
 				if (p1 == p1.ToLowerInvariant () && p1 == p1.ToUpperInvariant ()) {
-					msg = String.Format ("Prefer the use of: {0}('{1}'...);", call.Name, p1);
-					Runner.Report (method, ins, Severity.Medium, Confidence.High, msg);
+					Report (method, ins, Confidence.High, call, p1);
 				}
 				break;
 			}
@@ -134,6 +145,11 @@ namespace Gendarme.Rules.Performance {
 			Runner.Report (method, ins, Severity.Medium, Confidence.High, msg);
 		}
 
+		static bool CheckFirstParameterIsString (IMethodSignature method)
+		{
+			return (method.HasParameters && (method.Parameters [0].ParameterType.FullName == "System.String"));
+		}
+
 		public RuleResult CheckMethod (MethodDefinition method)
 		{
 			// rule apply only if the method has a body (e.g. p/invokes, icalls don't)
@@ -157,15 +173,13 @@ namespace Gendarme.Rules.Performance {
 				case "LastIndexOf":
 					// 3 overloads each - parameters are identical (between them)
 					// and the String (or Char) is always the first one
-					if (call.Parameters [0].ParameterType.FullName != "System.String")
-						continue;
-					CheckIndexOf (method, call, ins);
+					if (CheckFirstParameterIsString (call))
+						CheckIndexOf (method, call, ins);
 					break;
 				case "Replace":
-					if (call.Parameters [0].ParameterType.FullName != "System.String")
-						continue;
 					// both parameters needs to be length == 1
-					CheckReplace (method, ins);
+					if (CheckFirstParameterIsString (call))
+						CheckReplace (method, ins);
 					break;
 				}
 				// String.Split could be a candidate but it is unlikely we
