@@ -27,6 +27,7 @@
 //
 
 using System;
+using System.Linq;
 
 using Mono.Cecil;
 using Mono.Cecil.Cil;
@@ -41,10 +42,11 @@ namespace Gendarme.Rules.Performance {
 	/// <summary>
 	/// This rule looks for <c>InitOnly</c> fields (<c>readonly</c> in C#) that could be
 	/// turned into <c>Literal</c> (<c>const</c> in C#) because their value is known at
-	/// compile time. <c>Literal</c> fields don't need to be initialized (i.e. they wont
-	/// show up or add a static constructor to the type) resulting in less code and the 
-	/// value (not a reference to the field) will be directly used in the IL (but this
-	/// can be a double-edged sword).
+	/// compile time. <c>Literal</c> fields don't need to be initialized (i.e. they don't
+	/// force the compiler to add a static constructor to the type) resulting in less code and the 
+	/// value (not a reference to the field) will be directly used in the IL (which is OK
+	/// if the field has internal visibility, but is often problematic if the field is visible outside
+	/// the assembly).
 	/// </summary>
 	/// <example>
 	/// Bad example:
@@ -80,7 +82,7 @@ namespace Gendarme.Rules.Performance {
 				return RuleResult.DoesNotApply;
 
 			// get the static constructor
-			MethodDefinition cctor = type.Constructors.GetConstructor (true, Type.EmptyTypes);
+			MethodDefinition cctor = type.Methods.FirstOrDefault (m => m.IsConstructor && m.IsStatic);
 			if (cctor == null)
 				return RuleResult.DoesNotApply;
 
@@ -111,7 +113,7 @@ namespace Gendarme.Rules.Performance {
 				// and report constant stuff
 				if (Constant.Get (previous.OpCode.Code)) {
 					// adjust severity based on the field visibility and it's type
-					Severity s = (field.FieldType.FullName == Constants.String || !field.IsVisible ()) ?
+					Severity s = (field.FieldType.FullName == "System.String" || !field.IsVisible ()) ?
 						Severity.High : Severity.Medium;
 					Runner.Report (field, s, Confidence.Normal);
 				}
