@@ -42,9 +42,9 @@ namespace Gendarme {
 
 	public partial class Wizard : Form {
 
-		// used to call code asynchronously
-		delegate void MethodInvoker ();
-
+		// This is used for methods which we execute within a thread, but we don't
+		// execute the methods concurrently so we can use single thread.
+		[ThreadModel (ThreadModel.SingleThread)]
 		sealed class AssemblyInfo {
 			private DateTime timestamp;
 			private AssemblyDefinition definition;
@@ -62,6 +62,7 @@ namespace Gendarme {
 
 		private const string BaseUrl = "http://www.mono-project.com/";
 		private const string DefaultUrl = BaseUrl + "Gendarme";
+		private const string BugzillaUrl = "http://bugzilla.novell.com";
 
 		static Process process;
 
@@ -70,14 +71,14 @@ namespace Gendarme {
 		private GuiRunner runner;
 		private int counter;
 
-		private MethodInvoker assembly_loader;
+		private Action assembly_loader;
 		private IAsyncResult assemblies_loading;
+		private object loader_lock = new object ();
 
-		private MethodInvoker rule_loader;
+		private Action rule_loader;
 		private IAsyncResult rules_loading;
 
-		private MethodInvoker analyze;
-		private IAsyncResult analyzing;
+		private Action analyze;
 
 		private string html_report_filename;
 		private string xml_report_filename;
@@ -104,24 +105,19 @@ namespace Gendarme {
 			UpdatePageUI ();
 		}
 
+		[ThreadModel (ThreadModel.SingleThread)]
 		static void EndCallback (IAsyncResult result)
 		{
-			(result.AsyncState as MethodInvoker).EndInvoke (result);
-		}
-
-		static Process Process {
-			get {
-				if (process == null)
-					process = new Process ();
-				return process;
-			}
+			(result.AsyncState as Action).EndInvoke (result);
 		}
 
 		static void Open (string filename)
 		{
-			Process.StartInfo.Verb = "open";
-			Process.StartInfo.FileName = filename;
-			Process.Start ();
+			if (process == null)
+				process = new Process ();
+			process.StartInfo.Verb = "open";
+			process.StartInfo.FileName = filename;
+			process.Start ();
 		}
 
 		#region general wizard code
@@ -309,12 +305,32 @@ namespace Gendarme {
 
 		public void UpdateAssemblies ()
 		{
-			foreach (KeyValuePair<string,AssemblyInfo> kvp in assemblies) {
-				DateTime last_write = File.GetLastWriteTimeUtc (kvp.Key);
-				if ((kvp.Value.Definition == null) || (kvp.Value.Timestamp < last_write)) {
-					AssemblyInfo a = kvp.Value;
-					a.Timestamp = last_write;
-					a.Definition = AssemblyFactory.GetAssembly (kvp.Key);
+			if (IsDisposed)
+				throw new ObjectDisposedException (GetType ().Name);
+
+			// do not let the final check be done while we're still loading assemblies
+			lock (loader_lock) {
+				foreach (KeyValuePair<string,AssemblyInfo> kvp in assemblies) {
+					DateTime last_write = File.GetLastWriteTimeUtc (kvp.Key);
+					if ((kvp.Value.Definition == null) && (kvp.Value.Timestamp < last_write)) {
+						AssemblyInfo a = kvp.Value;
+						string filename = kvp.Key;
+						try {
+							a.Definition = AssemblyDefinition.ReadAssembly (filename, new ReaderParameters { AssemblyResolver = AssemblyResolver.Resolver });
+						}
+						catch (BadImageFormatException) {
+							// continue loading & analyzing assemblies
+							Runner.Warn ("Invalid image: " + filename);
+						}
+						catch (FileNotFoundException fnfe) {
+							// e.g. .netmodules
+							// continue loading & analyzing assemblies
+							Runner.Warn (fnfe.Message + " while loading " + filename);
+						}
+						finally {
+							a.Timestamp = last_write;
+						}
+					}
 				}
 			}
 		}
@@ -331,8 +347,10 @@ namespace Gendarme {
 			rules_count_label.Text = String.Format ("{0} rules are available.", Runner.Rules.Count);
 			if (rules_loading == null)
 				throw new InvalidOperationException ("rules_loading");
-			next_button.Enabled = rules_loading.IsCompleted;
-			rules_tree_view.Enabled = rules_loading.IsCompleted;
+
+			bool completed = rules_loading.IsCompleted;
+			next_button.Enabled = completed;
+			rules_tree_view.Enabled = completed;
 			rules_loading.AsyncWaitHandle.WaitOne ();
 			PopulateRules ();
 		}
@@ -352,7 +370,9 @@ namespace Gendarme {
 			rules_tree_view.AfterCheck -= RulesTreeViewAfterCheck;
 			foreach (IRule rule in Runner.Rules) {
 				TreeNode parent;
-				string name_space = rule.FullName.Substring (0, rule.FullName.Length - rule.Name.Length - 1);
+				string full_name = rule.FullName;
+				string name = rule.Name;
+				string name_space = full_name.Substring (0, full_name.Length - name.Length - 1);
 				if (!nodes.TryGetValue (name_space, out parent)) {
 					parent = new TreeNode (name_space);
 					parent.Checked = all_rules;
@@ -360,8 +380,8 @@ namespace Gendarme {
 					rules_tree_view.Nodes.Add (parent);
 				}
 
-				TreeNode node = new TreeNode (rule.Name);
-				node.Checked = all_rules || rules.Contains (rule.FullName);
+				TreeNode node = new TreeNode (name);
+				node.Checked = all_rules || rules.Contains (full_name);
 				node.Tag = rule;
 				node.ToolTipText = rule.Problem;
 				parent.Nodes.Add (node);
@@ -391,13 +411,14 @@ namespace Gendarme {
 		{
 			string url = null;
 
-			if (rules_tree_view.SelectedNode == null)
+			TreeNode selected = rules_tree_view.SelectedNode;
+			if (selected == null)
 				url = DefaultUrl;
 			else {
-				if (rules_tree_view.SelectedNode.Tag == null) {
-					url = BaseUrl + rules_tree_view.SelectedNode.Text;
+				if (selected.Tag == null) {
+					url = BaseUrl + selected.Text;
 				} else {
-					url = (rules_tree_view.SelectedNode.Tag as IRule).Uri.ToString ();
+					url = (selected.Tag as IRule).Uri.ToString ();
 				}
 			}
 
@@ -430,6 +451,7 @@ namespace Gendarme {
 			}
 		}
 
+		[ThreadModel (ThreadModel.SingleThread)]
 		private bool UpdateActiveRules ()
 		{
 			bool all_active = true;
@@ -512,7 +534,7 @@ namespace Gendarme {
 			assemblies_loading.AsyncWaitHandle.WaitOne ();
 			PrepareAnalyze ();
 			analyze = Analyze;
-			analyzing = analyze.BeginInvoke (EndCallback, analyze);
+			analyze.BeginInvoke (EndCallback, analyze);
 		}
 
 		private void PrepareAnalyze ()
@@ -530,12 +552,17 @@ namespace Gendarme {
 			Runner.Assemblies.Clear ();
 			foreach (KeyValuePair<string, AssemblyInfo> kvp in assemblies) {
 				// add assemblies references to runner
-				Runner.Assemblies.Add (kvp.Value.Definition);
+				AssemblyDefinition ad = kvp.Value.Definition;
+				// an invalid assembly (e.g. non-managed code) will be null at this stage
+				if (ad != null)
+					Runner.Assemblies.Add (ad);
 			}
 
 			progress_bar.Maximum = Runner.Assemblies.Count;
 		}
 
+		// TODO: Note that it isn't clear whether we can actually call gendarme from a thread...
+		[ThreadModel (ThreadModel.SingleThread)]
 		private void Analyze ()
 		{
 			counter = 0;
@@ -557,9 +584,8 @@ namespace Gendarme {
 			// activate rules based on user selection
 			UpdateActiveRules ();
 
-			Runner.Initialize ();
-			Runner.Run ();
-			Runner.TearDown ();
+			// Initialize / Run / TearDown
+			Runner.Execute ();
 
 			// reset rules as the user selected them (since some rules might have
 			// turned themselves off is they were not needed for the analysis) so
@@ -611,6 +637,32 @@ namespace Gendarme {
 				has_defects ? Runner.Defects.Count.ToString () : "no");
 			cancel_button.Text = "Close";
 			next_button.Enabled = false;
+
+			// display an error message and details if we encountered an exception during analysis
+			string error = Runner.Error;
+			bool has_errors = (error.Length > 0);
+			copy_paste_label.Visible = has_errors;
+			bugzilla_linklabel.Text = BugzillaUrl;
+			bugzilla_linklabel.Visible = has_errors;
+
+			string warnings = Runner.Warnings;
+			bool has_warnings = (warnings.Length > 0);
+
+			if (has_errors) {
+				unexpected_error_label.Text = "Results are incomplete due to an unexpected error!";
+				unexpected_error_label.Visible = true;
+				// give priority to errors before warnings
+				error_textbox.Text = error;
+				error_textbox.Visible = true;
+			} else if (has_warnings) {
+				unexpected_error_label.Text = "Results might be incomplete due to the following warnings!";
+				unexpected_error_label.Visible = true;
+				error_textbox.Text = warnings;
+				error_textbox.Visible = true;
+			} else {
+				unexpected_error_label.Visible = false;
+				error_textbox.Visible = false;
+			}
 		}
 
 		private static bool CouldCopyReport (ref string currentName, string fileName)
@@ -618,8 +670,15 @@ namespace Gendarme {
 			// if possible avoid re-creating the report (as it can 
 			// be a long operation) and simply copy the file
 			bool copy = (currentName != null);
-			if (copy)
-				File.Copy (currentName, fileName);
+			if (copy) {
+				try {
+					File.Copy (currentName, fileName);
+				}
+				catch (Exception) {
+					// too many things can go wrong when copying
+					copy = false;
+				}
+			}
 
 			currentName = fileName;
 			return copy;
@@ -653,10 +712,9 @@ namespace Gendarme {
 			if (save_file_dialog.ShowDialog () != DialogResult.OK)
 				return;
 
-			ResultWriter writer = GetSelectedWriter (save_file_dialog.FilterIndex, save_file_dialog.FileName);
-			if (writer != null) {
-				writer.Report ();
-				writer.Dispose ();
+			using (ResultWriter writer = GetSelectedWriter (save_file_dialog.FilterIndex, save_file_dialog.FileName)) {
+				if (writer != null)
+					writer.Report ();
 			}
 		}
 
@@ -672,6 +730,11 @@ namespace Gendarme {
 			Open (html_report_filename);
 		}
 
+		private void BugzillaLinkClick (object sender, LinkLabelLinkClickedEventArgs e)
+		{
+			Open (BugzillaUrl);
+		}
+
 		#endregion
 	}
 }
