@@ -24,9 +24,13 @@
 //
 
 using System;
+using System.Collections;
 using System.Collections.Generic;
+using System.Runtime.CompilerServices;
 using System.IO;
+using System.Linq;
 using System.Text;
+using System.Globalization;
 
 using Mono.Cecil;
 using Mono.Cecil.Cil;
@@ -60,7 +64,7 @@ namespace GuiCompare {
 			if (array != null)
 				return PrettyType (array.ElementType) + "[]";
 
-			var reference = type as ReferenceType;
+			var reference = type as ByReferenceType;
 			if (reference != null)
 				return PrettyType (reference.ElementType) + "&";
 
@@ -90,6 +94,30 @@ namespace GuiCompare {
 			return type.Name;
 		}
 		
+		public static string PrettyTypeDefinition (TypeDefinition td)
+		{
+			if (td.GenericParameters.Count > 0) {
+				int arity_start = td.Name.IndexOf ('`');
+				if (arity_start > 0) {
+					var sb = new StringBuilder (td.Name);
+					sb.Remove (arity_start, td.Name.Length - arity_start);
+					sb.Append ("<");
+					bool first_gp = true;
+					foreach (GenericParameter gp in td.GenericParameters) {
+						if (!first_gp)
+							sb.Append (',');
+						first_gp = false;
+						sb.Append (gp.Name);
+					}
+					sb.Append (">");
+				
+					return sb.ToString ();
+				}
+			}
+			
+			return td.Name;
+		}		
+		
 		// the corcompare xml output uses a different formatting than Cecil.
 		// Cecil uses / for nested classes, ala:
 		//  Namespace.Class/NestedClass
@@ -118,8 +146,8 @@ namespace GuiCompare {
 		{
 			if (interface_list != null) {
 				foreach (TypeReference ifc in GetInterfaces (fromDef)) {
-					TypeDefinition ifc_def = CecilUtils.Resolver.Resolve (ifc);
-					if (ifc_def.IsNotPublic)
+					TypeDefinition ifc_def = ifc.Resolve ();
+					if (ifc_def == null || ifc_def.IsNotPublic)
 						continue;
 
 					interface_list.Add (new CecilInterface (ifc));
@@ -127,21 +155,27 @@ namespace GuiCompare {
 			}
 
 			if (constructor_list != null) {
-				foreach (MethodDefinition md in fromDef.Constructors) {
+				foreach (MethodDefinition md in fromDef.Methods.Where (m => m.IsConstructor)) {
 					if (md.IsPrivate || md.IsAssembly)
 						continue;
 					constructor_list.Add (new CecilMethod (md));
 				}
 			}
 			if (method_list != null) {
-				foreach (MethodDefinition md in fromDef.Methods) {
+				foreach (MethodDefinition md in fromDef.Methods.Where (m => !m.IsConstructor)) {
 					if (md.IsSpecialName) {
 						if (!md.Name.StartsWith("op_"))
 							continue;
 					}
 
-					if (IsFinalizer (md))
-						continue;
+					if (IsFinalizer (md)) {
+						string name = md.DeclaringType.Name;
+						int arity = name.IndexOf ('`');
+						if (arity > 0)
+							name = name.Substring (0, arity);
+
+						md.Name = "~" + name;
+					}
 
 					if (md.IsPrivate || md.IsAssembly)
 						continue;
@@ -186,14 +220,9 @@ namespace GuiCompare {
 			}
 		}
 
-		static TypeDefinition Resolve (TypeReference type)
-		{
-			return CecilUtils.Resolver.Resolve (type);
-		}
-
 		static IEnumerable<TypeDefinition> WalkHierarchy (TypeReference type)
 		{
-			for (var definition = Resolve (type); definition != null; definition = GetBaseType (definition))
+			for (var definition = type.Resolve (); definition != null; definition = GetBaseType (definition))
 				yield return definition;
 		}
 
@@ -202,7 +231,7 @@ namespace GuiCompare {
 			if (type.BaseType == null)
 				return null;
 
-			return Resolve (type.BaseType);
+			return type.BaseType.Resolve ();
 		}
 
 		static IEnumerable<TypeReference> GetInterfaces (TypeReference type)
@@ -254,8 +283,7 @@ namespace GuiCompare {
 				else if (type_def.IsInterface) {
 					interface_list.Add (new CecilInterface (type_def));
 				}
-				else if (type_def.BaseType.FullName == "System.MulticastDelegate"
-				         || type_def.BaseType.FullName == "System.Delegate") {
+				else if (type_def.BaseType.FullName == "System.MulticastDelegate") {
 					delegate_list.Add (new CecilDelegate (type_def));
 				}
 				else {
@@ -268,11 +296,11 @@ namespace GuiCompare {
 		{
 			StringBuilder sb = new StringBuilder();
 			bool first = true;
-			foreach (object o in ca.ConstructorParameters) {
+			foreach (var argument in ca.ConstructorArguments) {
 				if (!first)
 					sb.Append (", ");
 				first = false;
-				sb.Append (o.ToString());
+				sb.Append (argument.Value.ToString());
 			}
 			
 			return sb.ToString();
@@ -289,22 +317,14 @@ namespace GuiCompare {
 			if (typedef.BaseType == null)
 				return false;
 			
-			return IsTODOAttribute (CecilUtils.Resolver.Resolve (typedef.BaseType));
-		}
-		
-		public static bool ShouldSkipAttribute (string name)
-		{
-			if (name == "System.Diagnostics.CodeAnalysis.SuppressMessageAttribute" || name == "System.Runtime.CompilerServices.CompilerGeneratedAttribute")
-				return true;
-			
-			return false;
+			return IsTODOAttribute (GetBaseType (typedef));
 		}
 		
 		public static List<CompNamed> GetCustomAttributes (ICustomAttributeProvider provider, List<string> todos)
 		{
 			List<CompNamed> rv = new List<CompNamed>();
 			foreach (CustomAttribute ca in provider.CustomAttributes) {
-				TypeDefinition resolved = CecilUtils.Resolver.Resolve (ca.Constructor.DeclaringType);
+				TypeDefinition resolved = ca.Constructor.DeclaringType.Resolve ();
 
 				if (resolved != null) {
 					if (IsTODOAttribute (resolved)) {
@@ -316,52 +336,84 @@ namespace GuiCompare {
 						continue;
 				}
 
-				if (!ShouldSkipAttribute (ca.Constructor.DeclaringType.FullName))
+				if (!MasterUtils.IsImplementationSpecificAttribute (ca.Constructor.DeclaringType.FullName))
 					rv.Add (new CecilAttribute (ca));
 			}
 			return rv;
 		}
-
 		
-		public static readonly AssemblyResolver Resolver = new AssemblyResolver();
+		public static List<CompGenericParameter> GetTypeParameters (IGenericParameterProvider provider)
+		{
+			if (provider.GenericParameters.Count == 0)
+				return null;
+				
+			var l = new List<CompGenericParameter> ();
+			
+			var gparameters = provider.GenericParameters;
+			foreach (GenericParameter gp in gparameters) {
+				l.Add (new CecilGenericParameter (gp));
+			}			
+			
+			return l;			
+		}
+		
+		public static List<CompParameter> GetParameters (IMethodSignature provider)
+		{
+			var l = new List<CompParameter> ();
+			foreach (ParameterDefinition pd in provider.Parameters)
+			{
+				l.Add (new CecilParameter (pd));
+			}
+					
+			return l;
+		}
+		
+		public static readonly IAssemblyResolver Resolver = new DefaultAssemblyResolver();
 	}
 
 	public class CecilAssembly : CompAssembly {
 		public CecilAssembly (string path)
-			: base (Path.GetFileName (path))
+			: base(Path.GetFileName (path))
 		{
-			Dictionary<string, Dictionary <string, TypeDefinition>> namespaces = new Dictionary<string, Dictionary <string, TypeDefinition>> ();
-
-			AssemblyDefinition assembly = AssemblyFactory.GetAssembly(path);
+			var namespaces = new Dictionary<string, Dictionary<string, TypeDefinition>> ();
 
+			var assembly = AssemblyDefinition.ReadAssembly (path, new ReaderParameters { AssemblyResolver = CecilUtils.Resolver });
+			
 			foreach (TypeDefinition t in assembly.MainModule.Types) {
 				if (t.Name == "<Module>")
 					continue;
-
+				
 				if (t.IsNotPublic)
 					continue;
-				
-				if (t.IsNested)
-					continue;
-				
+
 				if (t.IsSpecialName || t.IsRuntimeSpecialName)
 					continue;
 
 				if (CecilUtils.IsTODOAttribute (t))
 					continue;
 
-				if (!namespaces.ContainsKey (t.Namespace))
-					namespaces[t.Namespace] = new Dictionary <string, TypeDefinition> ();
+				Dictionary<string, TypeDefinition> ns;
 
-				namespaces[t.Namespace][t.Name] = t;
+				if (!namespaces.TryGetValue (t.Namespace, out ns)) {
+					ns = new Dictionary<string, TypeDefinition> ();
+					namespaces.Add (t.Namespace, ns);
+				}
+
+				ns[t.Name] = t;
 			}
 
-			namespace_list = new List<CompNamed>();
-			foreach (string ns_name in namespaces.Keys) {
+			namespace_list = new List<CompNamed> ();
+			foreach (string ns_name in namespaces.Keys)
 				namespace_list.Add (new CecilNamespace (ns_name, namespaces[ns_name]));
-			}
-			
+
 			attributes = CecilUtils.GetCustomAttributes (assembly, todos);
+
+			// TypeForwardedToAttributes are created by checking if assembly contains
+			// extern forwarder types and using them to construct fake custom attributes
+			foreach (ExportedType t in assembly.MainModule.ExportedTypes) {
+				if (t.IsForwarder)
+					attributes.Add (new PseudoCecilAttribute (t));
+			}
 		}
 
 		public override List<CompNamed> GetNamespaces()
@@ -387,7 +439,8 @@ namespace GuiCompare {
  			delegate_list = new List<CompNamed>();
 			interface_list = new List<CompNamed>();
 			struct_list = new List<CompNamed>();
-
+			MemberName = name;
+			
 			foreach (string type_name in type_mapping.Keys) {
 				TypeDefinition type_def = type_mapping[type_name];
 				if (type_def.IsNotPublic)
@@ -406,8 +459,7 @@ namespace GuiCompare {
 				else if (type_def.IsInterface) {
 					interface_list.Add (new CecilInterface (type_def));
 				}
-				else if ((type_def.BaseType != null && type_def.BaseType.FullName == "System.MulticastDelegate")
-				         || (type_def.BaseType != null && type_def.BaseType.FullName == "System.Delegate")) {
+				else if (type_def.BaseType != null && type_def.BaseType.FullName == "System.MulticastDelegate") {
 					delegate_list.Add (new CecilDelegate (type_def));
 				}
 				else {
@@ -453,6 +505,7 @@ namespace GuiCompare {
 			: base (type_def.Name)
 		{
 			this.type_def = type_def;
+			DisplayName = CecilUtils.PrettyTypeDefinition (type_def);
 			
 			interfaces = new List<CompNamed>();
 			constructors = new List<CompNamed>();
@@ -460,7 +513,10 @@ namespace GuiCompare {
 			properties = new List<CompNamed>();
 			fields = new List<CompNamed>();
 			events = new List<CompNamed>();
-
+			if (!type_def.IsNotPublic || type_def.IsPublic || type_def.IsNestedPublic || type_def.IsNestedFamily ||
+			    type_def.IsNestedFamilyAndAssembly || type_def.IsNestedFamilyOrAssembly)
+				MemberName = type_def.FullName;
+			
 			CecilUtils.PopulateMemberLists (type_def,
 			                                interfaces,
 			                                constructors,
@@ -470,8 +526,9 @@ namespace GuiCompare {
 			                                events);
 			
 			attributes = CecilUtils.GetCustomAttributes (type_def, todos);
+			tparams = CecilUtils.GetTypeParameters (type_def);
 		}
-		
+
 		public CecilInterface (TypeReference type_ref)
 			: base (CecilUtils.FormatTypeLikeCorCompare (type_ref))
 		{
@@ -483,6 +540,7 @@ namespace GuiCompare {
 			events = new List<CompNamed>();
 			
 			attributes = new List<CompNamed>();
+			tparams = new List<CompGenericParameter>();
 		}
 
 		public override string GetBaseType ()
@@ -525,6 +583,11 @@ namespace GuiCompare {
 			return attributes;
 		}
 		
+		public override List<CompGenericParameter> GetTypeParameters ()
+		{
+			return tparams;
+		}
+		
 		List<CompNamed> interfaces;
 		List<CompNamed> constructors;
 		List<CompNamed> methods;
@@ -532,6 +595,7 @@ namespace GuiCompare {
 		List<CompNamed> fields;
 		List<CompNamed> events;
 		List<CompNamed> attributes;
+		List<CompGenericParameter> tparams;
 		TypeDefinition type_def;
 	}
 
@@ -540,6 +604,18 @@ namespace GuiCompare {
 			: base (type_def.Name)
 		{
 			this.type_def = type_def;
+			DisplayName = CecilUtils.PrettyTypeDefinition (type_def);
+
+			if (!type_def.IsNotPublic || type_def.IsPublic || type_def.IsNestedPublic || type_def.IsNestedFamily ||
+			    type_def.IsNestedFamilyAndAssembly || type_def.IsNestedFamilyOrAssembly)
+				MemberName = type_def.FullName;
+				
+			attributes = CecilUtils.GetCustomAttributes (type_def, todos);				
+		}
+		
+		public override List<CompNamed> GetAttributes ()
+		{
+			return attributes;
 		}
 
 		public override string GetBaseType ()
@@ -547,7 +623,35 @@ namespace GuiCompare {
 			return type_def.BaseType == null ? null : CecilUtils.FormatTypeLikeCorCompare (type_def.BaseType);
 		}
 		
+		public override List<CompNamed> GetConstructors ()
+		{
+			List<CompNamed> l = new List<CompNamed> ();
+			foreach (MethodDefinition md in type_def.Methods) {
+				if (md.IsConstructor)
+					l.Add (new CecilMethod (md));
+			}
+			
+			return l;
+		}
+		
+		public override List<CompNamed> GetMethods ()
+		{
+			List<CompNamed> l = new List<CompNamed> ();
+			foreach (MethodDefinition md in type_def.Methods) {
+				if (!md.IsConstructor)
+					l.Add (new CecilMethod (md));
+			}
+			
+			return l;
+		}
+		
+		public override List<CompGenericParameter> GetTypeParameters ()
+		{
+			return CecilUtils.GetTypeParameters (type_def);
+		}
+		
 		TypeDefinition type_def;
+		List<CompNamed> attributes;
 	}
 
 	public class CecilEnum : CompEnum {
@@ -557,7 +661,10 @@ namespace GuiCompare {
 			this.type_def = type_def;
 
 			fields = new List<CompNamed>();
-
+			if (!type_def.IsNotPublic || type_def.IsPublic || type_def.IsNestedPublic || type_def.IsNestedFamily ||
+			    type_def.IsNestedFamilyAndAssembly || type_def.IsNestedFamilyOrAssembly)
+				MemberName = type_def.FullName;
+			
 			CecilUtils.PopulateMemberLists (type_def,
 						   null,
 						   null,
@@ -566,7 +673,7 @@ namespace GuiCompare {
 						   fields,
 						   null);
 			
-			attributes = CecilUtils.GetCustomAttributes (type_def, todos); 
+			attributes = CecilUtils.GetCustomAttributes (type_def, todos);
 		}
 
 		public override string GetBaseType ()
@@ -594,6 +701,7 @@ namespace GuiCompare {
 			: base (type_def.Name, type)
 		{
 			this.type_def = type_def;
+			DisplayName = CecilUtils.PrettyTypeDefinition (type_def);
 
 			nested_classes = new List<CompNamed>();
 			nested_enums = new List<CompNamed>();
@@ -615,6 +723,10 @@ namespace GuiCompare {
 			fields = new List<CompNamed>();
 			events = new List<CompNamed>();
 
+			if (!type_def.IsNotPublic || type_def.IsPublic || type_def.IsNestedPublic || type_def.IsNestedFamily ||
+			    type_def.IsNestedFamilyAndAssembly || type_def.IsNestedFamilyOrAssembly)
+				MemberName = type_def.FullName;
+			
 			CecilUtils.PopulateMemberLists (type_def,
 			                           interfaces,
 			                           constructors,
@@ -630,6 +742,9 @@ namespace GuiCompare {
 		{
 			return type_def.BaseType == null ? null : CecilUtils.FormatTypeLikeCorCompare (type_def.BaseType);
 		}
+		
+		public override bool IsAbstract { get { return type_def.IsAbstract; } }
+		public override bool IsSealed { get { return type_def.IsSealed; } }
 
 		public override List<CompNamed> GetInterfaces ()
 		{
@@ -690,6 +805,11 @@ namespace GuiCompare {
 		{
 			return nested_delegates;
 		}
+		
+		public override List<CompGenericParameter> GetTypeParameters ()
+		{
+			return CecilUtils.GetTypeParameters (type_def);
+		}
 
 		TypeDefinition type_def;
 		List<CompNamed> nested_classes;
@@ -713,6 +833,11 @@ namespace GuiCompare {
 		{
 			this.field_def = field_def;
 			this.attributes = CecilUtils.GetCustomAttributes (field_def, todos);
+			if (field_def.IsPublic || field_def.IsFamily || field_def.IsFamilyAndAssembly || field_def.IsFamilyOrAssembly) {
+				TypeDefinition declType = field_def.DeclaringType;
+				if (declType != null)
+					MemberName = declType.FullName + "." + field_def.Name;
+			}
 		}
 
 		public override string GetMemberType ()
@@ -745,8 +870,13 @@ namespace GuiCompare {
 
 		public override string GetLiteralValue ()
 		{
-			if (field_def.IsLiteral)
-				return field_def.Constant.ToString();
+			if (field_def.IsLiteral && field_def.Constant != null) {
+				if (field_def.Constant is char)
+					return string.Format (CultureInfo.InvariantCulture, (char) field_def.Constant < 0x80 ? "\\x{0:X02}" : "\\u{0:X04}", (int)(char) field_def.Constant);
+				
+				return Convert.ToString (field_def.Constant, CultureInfo.InvariantCulture);
+			}
+			
 			return null;
 		}
 		
@@ -761,6 +891,11 @@ namespace GuiCompare {
 			this.method_def = method_def;
 			this.attributes = CecilUtils.GetCustomAttributes (method_def, todos);
 			DisplayName = FormatName (method_def, true);
+			if (method_def.IsFamily || method_def.IsFamilyAndAssembly || method_def.IsFamilyOrAssembly || method_def.IsPublic) {
+				TypeReference declType = method_def.DeclaringType;
+				if (declType != null)
+					MemberName = declType.FullName + "." + method_def.Name;
+			}
 		}
 
 		public override string GetMemberType ()
@@ -768,7 +903,7 @@ namespace GuiCompare {
 			if (method_def.IsConstructor)
 				return null;
 			
-			return CecilUtils.FormatTypeLikeCorCompare (method_def.ReturnType.ReturnType);
+			return CecilUtils.FormatTypeLikeCorCompare (method_def.ReturnType);
 		}
 
 		public override bool ThrowsNotImplementedException ()
@@ -805,13 +940,23 @@ namespace GuiCompare {
 			return attributes;
 		}
 		
+		public override List<CompGenericParameter> GetTypeParameters ()
+		{
+			return CecilUtils.GetTypeParameters (method_def);
+		}
+		
+		public override List<CompParameter> GetParameters ()
+		{
+			return CecilUtils.GetParameters (method_def);
+		}
+		
 		static string FormatName (MethodDefinition method_def, bool beautify)
 		{
 			StringBuilder sb = new StringBuilder ();
 			if (!method_def.IsConstructor)
 				sb.Append (beautify
-				           ? CecilUtils.PrettyType (method_def.ReturnType.ReturnType)
-				           : CecilUtils.FormatTypeLikeCorCompare (method_def.ReturnType.ReturnType));
+				           ? CecilUtils.PrettyType (method_def.ReturnType)
+				           : CecilUtils.FormatTypeLikeCorCompare (method_def.ReturnType));
 			sb.Append (" ");
 			if (beautify) {
 				if (method_def.IsSpecialName && method_def.Name.StartsWith ("op_")) {
@@ -840,6 +985,8 @@ namespace GuiCompare {
 					case "op_OnesComplement": sb.Append ("operator ~"); break;
 					case "op_True": sb.Append ("operator true"); break;
 					case "op_False": sb.Append ("operator false"); break;
+					case "op_LeftShift": sb.Append ("operator <<"); break;
+					case "op_RightShift": sb.Append ("operator >>"); break;
 					default: Console.WriteLine ("unhandled operator named {0}", method_def.Name); sb.Append (method_def.Name); break;
 					}
 				}
@@ -896,6 +1043,22 @@ namespace GuiCompare {
 			this.pd = pd;
 			this.attributes = CecilUtils.GetCustomAttributes (pd, todos);
 			this.DisplayName = FormatName (pd, true);
+
+			MethodDefinition getMethod = pd.GetMethod, setMethod = pd.SetMethod;			
+			if (getMethod != null || setMethod != null) {
+				bool interesting = false;
+
+				if (getMethod != null && (getMethod.IsPublic || getMethod.IsFamily || getMethod.IsFamilyAndAssembly || getMethod.IsFamilyOrAssembly))
+					interesting = true;
+				else if (setMethod != null && (setMethod.IsPublic || setMethod.IsFamily || setMethod.IsFamilyAndAssembly || setMethod.IsFamilyOrAssembly))
+					interesting = true;
+
+				if (interesting) {
+					TypeDefinition declType = pd.DeclaringType;
+					if (declType != null)
+						MemberName = declType.FullName + "." + pd.Name;
+				}
+			}
 		}
 
 		public override string GetMemberType()
@@ -974,6 +1137,22 @@ namespace GuiCompare {
 		{
 			this.ed = ed;
 			this.attributes = CecilUtils.GetCustomAttributes (ed, todos);
+
+			MethodDefinition addMethod = ed.AddMethod, removeMethod = ed.RemoveMethod;
+			if (addMethod != null || removeMethod != null) {
+				bool interesting = false;
+
+				if (addMethod != null && (addMethod.IsPublic || addMethod.IsFamily || addMethod.IsFamilyAndAssembly || addMethod.IsFamilyOrAssembly))
+					interesting = true;
+				else if (removeMethod != null && (removeMethod.IsPublic || removeMethod.IsFamily || removeMethod.IsFamilyAndAssembly || removeMethod.IsFamilyOrAssembly))
+					interesting = true;
+
+				if (interesting) {
+					TypeDefinition declType = ed.DeclaringType;
+					if (declType != null)
+						MemberName = declType.FullName + "." + ed.Name;
+				}
+			}
 		}
 
 		public override string GetMemberType()
@@ -1000,9 +1179,110 @@ namespace GuiCompare {
 		public CecilAttribute (CustomAttribute ca)
 			: base (ca.Constructor.DeclaringType.FullName)
 		{
-			this.ca = ca;
+			var sb = new StringBuilder ("[" + ca.Constructor.DeclaringType.FullName);
+			bool first = true;
+			
+			var cargs = ca.ConstructorArguments;
+			if (cargs != null && cargs.Count > 0) {
+				foreach (var argument in cargs) {
+					if (first) {
+						sb.Append (" (");
+						first = false;
+					} else
+						sb.Append (", ");
+
+					sb.Append (FormatValue (argument.Value));
+				}
+				
+			}
+
+			var properties = ca.Properties;
+			if (properties != null && properties.Count > 0) {
+				foreach (var namedArg in properties) {
+					if (first) {
+						sb.Append (" (");
+						first = false;
+					} else
+						sb.Append (", ");
+					
+					sb.AppendFormat ("{0}={1}", namedArg.Name, FormatValue (namedArg.Argument.Value));
+				}
+			}
+			
+			if (!first)
+				sb.Append (')');
+			sb.Append ("]");
+			
+			ExtraInfo = sb.ToString ();
+		}
+
+		string FormatValue (object o)
+		{
+			if (o == null)
+				return "null";
+
+			if (o is string)
+				return "\"" + o + "\"";
+
+			if (o is bool)
+				return o.ToString ().ToLowerInvariant ();
+			
+			return o.ToString ();
+		}
+	}
+
+	public class PseudoCecilAttribute : CompAttribute
+	{
+		public PseudoCecilAttribute (ExportedType type)
+			: base (typeof (TypeForwardedToAttribute).FullName)
+		{
+			ExtraInfo = "[assembly: TypeForwardedToAttribute (typeof (" + type.ToString () + "))]";
+		}
+	}
+	
+	public class CecilGenericParameter : CompGenericParameter
+	{
+		List<CompNamed> attributes;	
+		IList<TypeReference> constraints;
+		
+		public CecilGenericParameter (GenericParameter gp)
+			: base (gp.Name, gp.Attributes)
+		{
+			attributes = CecilUtils.GetCustomAttributes (gp, todos);
+			
+			var constraints = gp.Constraints;
+			if (constraints.Count == 0)
+				return;
+			
+			// TODO: finish constraints loading
+			this.constraints = constraints;
 		}
 		
-		CustomAttribute ca;
+		public override bool HasConstraints {
+			get {
+				return constraints != null;
+			}
+		}
+		
+		public override List<CompNamed> GetAttributes ()
+		{
+			return attributes;
+		}
+	}
+	
+	public class CecilParameter : CompParameter
+	{
+		List<CompNamed> attributes;
+		
+		public CecilParameter (ParameterDefinition pd)
+			: base (pd.Name, CecilUtils.FormatTypeLikeCorCompare (pd.ParameterType), pd.IsOptional)
+		{
+			attributes = CecilUtils.GetCustomAttributes (pd, todos);
+		}
+		
+		public override List<CompNamed> GetAttributes ()
+		{
+			return attributes;
+		}
 	}
 }
